// Copyright (c) 2017-2018 Hartmut Kaiser
//
// Distributed under the Boost Software License, Version 1.0. (See accompanying
// file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include <phylanx/config.hpp>
#include <phylanx/ast/generate_ast.hpp>
#include <phylanx/ast/node.hpp>
#include <phylanx/execution_tree/compile.hpp>
#include <phylanx/execution_tree/compiler/compiler.hpp>
#include <phylanx/execution_tree/compiler_component.hpp>
#include <phylanx/execution_tree/primitives/base_primitive.hpp>

#include <hpx/format.hpp>
#include <hpx/include/naming.hpp>

#include <algorithm>
#include <atomic>
#include <cstddef>
#include <string>
#include <utility>
#include <vector>

namespace phylanx { namespace execution_tree
{
    ///////////////////////////////////////////////////////////////////////////
    namespace detail
    {
        compiler::function compile(std::string const& name,
            ast::expression const& expr, compiler::function_list& snippets,
            compiler::environment& env, hpx::id_type const& default_locality)
        {
            ++snippets.compile_id_;
            return compiler::compile(name, expr, snippets, env,
                compiler::generate_patterns(), default_locality);
        }

        compiler::function compile(std::string const& name,
            ast::expression const& expr, compiler::function_list& snippets,
            hpx::id_type const& default_locality)
        {
            compiler::environment env =
                compiler::default_environment(default_locality);

            ++snippets.compile_id_;
            return compiler::compile(name, expr, snippets, env,
                compiler::generate_patterns(), default_locality);
        }

        compiler::function compile(std::string const& name,
            ast::expression const& expr, compiler::function_list& snippets,
            compiler::environment& env,
            compiler::expression_pattern_list const& patterns,
            hpx::id_type const& default_locality)
        {
            ++snippets.compile_id_;
            return compiler::compile(
                name, expr, snippets, env, patterns, default_locality);
        }
    }

    ///////////////////////////////////////////////////////////////////////////
    compiler::entry_point const& compile(std::string const& name,
        std::string const& func_name, std::vector<ast::expression> const& exprs,
        compiler::function_list& snippets, compiler::environment& env,
        compiler::expression_pattern_list const& patterns,
        hpx::id_type const& default_locality)
    {
        compiler::entry_point entry_point(func_name, name);

        for (auto const& expr : exprs)
        {
            // always keep objects alive that are generated by the compiler
            entry_point.add_entry_point(detail::compile(
                name, expr, snippets, env, patterns, default_locality));
        }

        // always return the last of all generated compiler-functions
        return snippets.program_.add_entry_point(std::move(entry_point));
    }

    compiler::entry_point const& compile(std::string const& name,
        std::string const& func_name, std::vector<ast::expression> const& exprs,
        compiler::function_list& snippets, compiler::environment& env,
        hpx::id_type const& default_locality)
    {
        compiler::entry_point entry_point(func_name, name);

        for (auto const& expr : exprs)
        {
            // always keep objects alive that are generated by the compiler
            entry_point.add_entry_point(
                detail::compile(name, expr, snippets, env, default_locality));
        }

        // always return the last of all generated compiler-functions
        return snippets.program_.add_entry_point(std::move(entry_point));
    }

    ////////////////////////////////////////////////////////////////////////////
    namespace detail
    {
        static std::atomic<std::size_t> function_counter(0);
        static std::string generate_unique_function_name()
        {
            return hpx::util::format("function_{}", ++function_counter);
        }
    }

    compiler::entry_point const& compile(std::string const& name,
        std::vector<ast::expression> const& exprs,
        compiler::function_list& snippets, compiler::environment& env,
        hpx::id_type const& default_locality)
    {
        return compile(name, detail::generate_unique_function_name(), exprs,
            snippets, env, default_locality);
    }

    compiler::entry_point const& compile(std::string const& name,
        std::string const& expr, compiler::function_list& snippets,
        compiler::environment& env, hpx::id_type const& default_locality)
    {
        return compile(name, detail::generate_unique_function_name(),
            ast::generate_ast(expr), snippets, env, default_locality);
    }

    compiler::entry_point const& compile(std::string const& name,
        std::string const& func_name, std::string const& expr,
        compiler::function_list& snippets, compiler::environment& env,
        hpx::id_type const& default_locality)
    {
        return compile(name, func_name, ast::generate_ast(expr), snippets, env,
            default_locality);
    }

    compiler::entry_point const& compile(std::string const& name,
        std::string const& func_name, std::vector<ast::expression> const& exprs,
        compiler::function_list& snippets, hpx::id_type const& default_locality)
    {
        compiler::environment env =
            compiler::default_environment(default_locality);

        compiler::entry_point entry_point(func_name, name);

        for (auto const& expr : exprs)
        {
            // always keep objects alive that are generated by the compiler
            entry_point.add_entry_point(
                detail::compile(name, expr, snippets, env, default_locality));
        }

        // always return the last of all generated compiler-functions
        return snippets.program_.add_entry_point(std::move(entry_point));
    }

    compiler::entry_point const& compile(std::string const& name,
        std::string const& expr, compiler::function_list& snippets,
        hpx::id_type const& default_locality)
    {
        return compile(name, detail::generate_unique_function_name(),
            ast::generate_ast(expr), snippets, default_locality);
    }

    compiler::entry_point const& compile(std::string const& name,
        std::vector<ast::expression> const& exprs,
        compiler::function_list& snippets, hpx::id_type const& default_locality)
    {
        return compile(name, detail::generate_unique_function_name(), exprs,
            snippets, default_locality);
    }

    compiler::entry_point const& compile(std::string const& name,
        std::string const& func_name, std::string const& expr,
        compiler::function_list& snippets, hpx::id_type const& default_locality)
    {
        return compile(name, func_name, ast::generate_ast(expr), snippets,
            default_locality);
    }

    ///////////////////////////////////////////////////////////////////////////
    compiler::entry_point const& compile(
        std::vector<ast::expression> const& exprs,
        compiler::function_list& snippets, compiler::environment& env,
        hpx::id_type const& default_locality)
    {
        return compile("<unknown>", detail::generate_unique_function_name(),
            exprs, snippets, env, default_locality);
    }

    compiler::entry_point const& compile(std::string const& expr,
        compiler::function_list& snippets, compiler::environment& env,
        hpx::id_type const& default_locality)
    {
        return compile("<unknown>", detail::generate_unique_function_name(),
            ast::generate_ast(expr), snippets, env, default_locality);
    }

    compiler::entry_point const& compile(
        std::vector<ast::expression> const& exprs,
        compiler::function_list& snippets, hpx::id_type const& default_locality)
    {
        return compile("<unknown>", detail::generate_unique_function_name(),
            exprs, snippets, default_locality);
    }

    compiler::entry_point const& compile(std::string const& expr,
        compiler::function_list& snippets, hpx::id_type const& default_locality)
    {
        return compile("<unknown>", detail::generate_unique_function_name(),
            ast::generate_ast(expr), snippets, default_locality);
    }

    ///////////////////////////////////////////////////////////////////////////
    /// Add the given variable to the compilation environment
    compiler::function define_variable(std::string const& codename,
        compiler::primitive_name_parts const& name_parts,
        compiler::function_list& snippets, compiler::environment& env,
        primitive_argument_type body, hpx::id_type const& default_locality)
    {
        return compiler::define_variable(
            codename, name_parts, snippets, env, body, default_locality);
    }

    ///////////////////////////////////////////////////////////////////////////
    /// Bind the given arguments to the function
    compiler::function bind_arguments(std::string const& codename,
        std::string const& func_name, compiler::function_list& snippets,
        primitive_argument_type func, primitive_arguments_type args,
        hpx::id_type const& default_locality)
    {
        return compiler::bind_arguments(codename, func_name, snippets,
            std::move(func), std::move(args), default_locality);
    }

    ///////////////////////////////////////////////////////////////////////////
    namespace compiler { namespace detail
    {
        void insert_pattern(expression_pattern_list& result,
            std::string pattern, match_pattern_type const& p,
            std::string const& suffix);
    }}

    void add_patterns(compiler::expression_pattern_list& patterns,
        match_pattern_type const& match)
    {
        for (auto const& p : match.patterns_)
        {
            compiler::detail::insert_pattern(patterns, p, match, "");
        }
    }
}}

